/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2016                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/
/*! \file trimesh_base.cpp
\ingroup code_sample

\brief the minimal example of using the lib

This file contain a minimal example of the library, showing how to load a mesh and how to compute per vertex normals on it.

*/

#include<vcg/complex/complex.h>
#include<vcg/complex/algorithms/update/color.h>
#include<vcg/complex/algorithms/point_sampling.h>
#include<vcg/complex/algorithms/clean.h>
#include<wrap/io_trimesh/import_off.h>
#include<wrap/io_trimesh/import_obj.h>

#include<wrap/io_trimesh/export_off.h>
#include<wrap/io_trimesh/export_ply.h>
#include<vcg/complex/algorithms/clean.h>

using namespace vcg;

class MyVertex; class MyEdge; class MyFace;
struct MyUsedTypes : public vcg::UsedTypes<vcg::Use<MyVertex>   ::AsVertexType,
                                           vcg::Use<MyEdge>     ::AsEdgeType,
                                           vcg::Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public vcg::Vertex< MyUsedTypes, vcg::vertex::Coord3f, vcg::vertex::Normal3f, vcg::vertex::Color4b, vcg::vertex::Qualityf, vcg::vertex::BitFlags  >{};
class MyFace    : public vcg::Face<   MyUsedTypes, vcg::face::FFAdj,  vcg::face::VertexRef, vcg::face::Color4b, vcg::face::Normal3f, vcg::face::BitFlags > {};
class MyEdge    : public vcg::Edge<   MyUsedTypes, vcg::edge::VertexRef, vcg::edge::EEAdj, vcg::edge::BitFlags > {};

class MyMesh    : public vcg::tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace> , std::vector<MyEdge>  > {};

int main( int argc, char **argv )
{
  if(argc<2)
  {
    printf("Usage trimesh_base <meshfilename.obj>\n");
    return -1;
  }
  /*!
    */
  MyMesh m, edgeMesh;
  srand(time(NULL));

  if(vcg::tri::io::ImporterOFF<MyMesh>::Open(m,argv[1])!=vcg::tri::io::ImporterOFF<MyMesh>::NoError)
  {
    printf("Error reading file  %s\n",argv[1]);
    exit(0);
  }

  tri::RequirePerVertexNormal(m);
  tri::RequirePerFaceNormal(m);
  tri::UpdateNormal<MyMesh>::PerVertexNormalized(m);
  tri::UpdateNormal<MyMesh>::PerFaceNormalized(m);
  tri::UpdateTopology<MyMesh>::FaceFace(m);
  tri::UpdateColor<MyMesh>::PerVertexConstant(m,vcg::Color4b::White);
  tri::UpdateQuality<MyMesh>::VertexConstant(m,0.0f);

  printf("Input mesh  vn:%i fn:%i\n",m.VN(),m.FN());
  vcg::tri::io::ExporterPLY<MyMesh>::Save(m,"input.ply", vcg::tri::io::Mask::IOM_VERTCOLOR|tri::io::Mask::IOM_VERTQUALITY); 

 // we search for all the edges where the difference between the normals of the two incident faces 
 // is greater than a given threshold
  float thrDeg = 30.0f;
  float cosThr = cosf( math::ToRad(thrDeg) );
  for(MyMesh::FaceIterator fi = m.face.begin(); fi != m.face.end(); ++fi)
  {
    MyMesh::FaceType &f = *fi;
    for(int i = 0; i < 3; i++)
    {
      float dot = f.N() * f.FFp(i)->N();
      if((dot < cosThr) && (f.FFp(i) > &f) )
      {
        f.V(i)->C() = vcg::Color4b::Red;
        f.V1(i)->C() = vcg::Color4b::Red;
        
        f.V0(i)->Q() +=1.0f;
        f.V1(i)->Q() +=1.0f;

        tri::Allocator<MyMesh>::AddEdge(edgeMesh, f.P0(i), f.P1(i));
      }
    }
  }

  for(MyMesh::VertexIterator vi = m.vert.begin(); vi != m.vert.end(); ++vi)
  {
    MyMesh::VertexType &v = *vi;
    if(v.Q() >= 3.0f)
      v.C() = vcg::Color4b::Blue;
  }

  tri::Clean<MyMesh>::RemoveDuplicateVertex(edgeMesh);
  tri::Allocator<MyMesh>::CompactEveryVector(edgeMesh);

  vcg::tri::io::ExporterPLY<MyMesh>::Save(m,"vertcol.ply", vcg::tri::io::Mask::IOM_VERTCOLOR); 
  vcg::tri::io::ExporterPLY<MyMesh>::Save(edgeMesh,"edge.ply", vcg::tri::io::Mask::IOM_EDGEINDEX);

  // Montecarlo sampling of the edge mesh
  tri::TrivialSampler<MyMesh> ts;
  
  tri::UpdateBounding<MyMesh>::Box(edgeMesh); 
  float radius = m.bbox.Diag() * 0.01f;

  tri::Clean<MyMesh>::SelectNonManifoldVertexOnEdgeMesh(edgeMesh);

  for(MyMesh::EdgeIterator ei = edgeMesh.edge.begin(); ei != edgeMesh.edge.end(); ++ei)
  {
    MyMesh::EdgeType &e = *ei;
    if(e.V(0)->IsS() || e.V(1)->IsS() ) tri::Allocator<MyMesh>::DeleteEdge(edgeMesh, e);  
  }

  for(MyMesh::VertexIterator vi = edgeMesh.vert.begin(); vi != edgeMesh.vert.end(); ++vi)
  {
   if( (*vi).IsS()) tri::Allocator<MyMesh>::DeleteVertex(edgeMesh, *vi);    
  }
  tri::Allocator<MyMesh>::CompactEveryVector(edgeMesh);
  vcg::tri::io::ExporterPLY<MyMesh>::Save(edgeMesh,"edgeManif.ply", vcg::tri::io::Mask::IOM_EDGEINDEX);

  tri::SurfaceSampling<MyMesh, tri::TrivialSampler < MyMesh> >::EdgeMeshUniform(edgeMesh, ts, radius); 

  printf("Sampled %i points\n",ts.SampleVec().size());

  MyMesh samples;
  for(auto it = ts.SampleVec().begin(); it != ts.SampleVec().end(); ++it)
  {
    tri::Allocator<MyMesh>::AddVertex(samples, *it);
  }
  vcg::tri::io::ExporterPLY<MyMesh>::Save(samples,"SamplePointsOnEdges.ply");

  // Todo Build a Poisson surface sampling function that sample accurately both the feature vertex and 
  // the feature edges of a given mesh. 
  // Things to do:
  // 1) Implement Uniform edge sampling strategy that works correctly with respect to non 1-manifold edges
  // 2) Implement Poisson surface sampling strategy that can take in input a set 
  //   of pre sampled vertices over the feature edges and vertices that must be removed 
  //   from the montecarlo samples before the other ones 

  // To implement 1 we write a SplitNon1ManifoldVertex function that 
  // split the non 1-manifold vertex in an edge mesh
	
  return 0;
}
